/*
 * Copyright (c) 2024.  Das Land Schleswig-Holstein vertreten durch den Ministerpräsidenten des Landes
 * Schleswig-Holstein Staatskanzlei Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.postfach.muk;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static de.ozgcloud.nachrichten.postfach.muk.transfer.ElsterTransferRemoteService.*;
import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.TestPropertySource;

import com.github.tomakehurst.wiremock.WireMockServer;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.nachrichten.postfach.muk.transfer.ElsterRestClient;
import de.ozgcloud.nachrichten.postfach.muk.transfer.ElsterTransferConfiguration;
import de.ozgcloud.nachrichten.postfach.muk.transfer.ElsterTransferRemoteService;

@SpringBootTest(
  classes = { MukPostfachTestConfiguration.class, MukPostfachRemoteService.class, ElsterTransferRemoteService.class,
	ElsterTransferConfiguration.class, ElsterRestClient.class, MukPostfachProperties.class })
@TestPropertySource(properties = { "ozgcloud.muk.server=localhost", "ozgcloud.muk.user-name=user", "ozgcloud.muk.password=xxxx" })
@ITCase
public class MukPostfachRemoteServiceITCase {
	private static final String DOWNLOAD_URL = "/rest/abholauftrag/12345/download/be123456789011122233301234567890?anhang=3";
	private static WireMockServer wireMockServer;
	@Autowired
	private MukPostfachRemoteService mukPostfachRemoteService;
	@Autowired
	private MukPostfachProperties properties;

	@BeforeAll
	static void startWireMockServer() {
		wireMockServer = new WireMockServer();
		wireMockServer.start();
	}

	@AfterAll
	static void stopWireMockServer() {
		wireMockServer.stop();
	}

	@Nested
	class TestSendingNachricht {

		@BeforeEach
		void setUp() {
			wireMockServer.resetAll();
			var url = wireMockServer.baseUrl();
			properties.setServer(url);

			stubFor(post(BEREITSTELLUNG_UPLOAD_ENDPUNKT)
			  .willReturn(
				aResponse()
				  .withHeader("Content-Type", "application/json")
				  .withBody(TestUtils.loadTextFile("UploadInfo.json"))));

			stubFor(post(BEREITSTELLUNGS_ENDPUNKT)
			  .willReturn(
				aResponse()
				  .withHeader("Content-Type", "application/json")
				  .withBody(TestUtils.loadTextFile("AuftragAngelegt.json"))));
		}

		@Test
		void shouldSendNachricht() {
			mukPostfachRemoteService.sendMessage(PostfachNachrichtTestFactory.create());

			verify(postRequestedFor(urlEqualTo(BEREITSTELLUNGS_ENDPUNKT))
			  .withHeader("Content-Type", equalTo("application/json")));
		}

		@Test
		void shouldUploadAttachment() {
			mukPostfachRemoteService.sendMessage(PostfachNachrichtTestFactory.create());

			verify(postRequestedFor(urlEqualTo(BEREITSTELLUNG_UPLOAD_ENDPUNKT)));
		}
	}

	@Nested
	class TestLoadingReplies {

		@BeforeEach
		void setUp() {
			wireMockServer.resetAll();
			var url = wireMockServer.baseUrl();
			properties.setServer(url);

			stubFor(get(ANTWORTEN_ENDPOINT)
			  .willReturn(
				aResponse()
				  .withHeader("Content-Type", "application/json")
				  .withBody(TestUtils.loadTextFile("AntwortDetailList.json"))));

			stubFor(get(DOWNLOAD_URL)
			  .willReturn(
				aResponse()
				  .withHeader("Content-Type", "application/octet-stream")
				  .withBody(TestUtils.contentStreamToByteArray(TestUtils.loadFile("Antwort-1.pdf")))));
		}

		@Test
		@Disabled("Disabled because loading of answers is disabled")
		void shouldNotLoadAntwortenBecauseNoVorgangIdInDateibezeichnung() {
			var antworten = mukPostfachRemoteService.getAllMessages();

			verify(getRequestedFor(urlEqualTo(ANTWORTEN_ENDPOINT)));
			assertThat(antworten).isEmpty();
		}

		@Test
		@Disabled("Need to be fixed when new API is available")
		void shouldLoadAntworten() {
			var antworten = mukPostfachRemoteService.getAllMessages();

			verify(getRequestedFor(urlEqualTo(ANTWORTEN_ENDPOINT)));
			assertThat(antworten).isNotEmpty();
		}

		@Test
		@Disabled("Needs to be reimplemented when the new File API is available")
		void shouldDownloadAntwort() {
			var antworten = mukPostfachRemoteService.getAllMessages().toList();

			verify(getRequestedFor(urlEqualTo(DOWNLOAD_URL)));
			assertThat(antworten).hasSize(1);
		}
	}
}
