/*
 * Copyright (c) 2024.  Das Land Schleswig-Holstein vertreten durch den Ministerpräsidenten des Landes
 * Schleswig-Holstein Staatskanzlei Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.postfach.muk.transfer;

import static de.ozgcloud.apilib.common.command.grpc.OzgCloudCommandTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.muk.BereitstellungAuftragTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.muk.PostfachNachrichtTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.muk.transfer.AntwortDetailsTestFactory.*;
import static de.ozgcloud.nachrichten.postfach.muk.transfer.ElsterTransferRemoteService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.util.UUID;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.muk.elster.transfer.BereitstellungAnhangWithUploadId;
import de.ozgcloud.muk.elster.transfer.BereitstellungAuftragNeuV3;
import de.ozgcloud.muk.elster.transfer.BereitstellungAuftragStatus;
import de.ozgcloud.muk.elster.transfer.BereitstellungInfo;
import de.ozgcloud.muk.elster.transfer.DownloadAnhangDetails;
import de.ozgcloud.muk.elster.transfer.UploadInfo;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.muk.BereitstellungAnhangTestFactory;
import de.ozgcloud.nachrichten.postfach.muk.BereitstellungAuftragTestFactory;
import de.ozgcloud.nachrichten.postfach.muk.MukPostfachProperties;
import de.ozgcloud.nachrichten.postfach.muk.PostfachNachrichtTestFactory;

class ElsterTransferRemoteServiceTest {
	@Spy
	@InjectMocks
	ElsterTransferRemoteService elsterTransferService;
	@Mock
	private MukPostfachProperties postfachProperties;
	@Mock
	private ElsterRestClient restClient;
	@Mock
	private MukPostfachMessageMapper mapper;
	@Mock
	private MukOzgCloudFileMapper fileMapper;
	@Mock
	private AttachmentService attachmentService;

	@Nested
	class TestCreatingBereitstellungsauftrag {
		BereitstellungAuftragNeuV3 auftragNeuV3 = BereitstellungAuftragTestFactory.createAuftrag();
		UploadInfo uploadInfo = new UploadInfo().uploadId(UUID.randomUUID());
		PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

		@Test
		void shouldHaveBescheidDatum() throws IOException {
			setup(MukPostfachMessageMapper.DATENART_BESCHEID);

			var filledAuftrag = elsterTransferService.createBereitstellungsAuftrag(nachricht);

			assertThat(filledAuftrag.getBescheidDatum()).isEqualTo(BESCHEID_DATUM);
		}

		@Test
		void shouldNotSetBescheidDatumForMitteilung() throws IOException {
			setup(MukPostfachMessageMapper.DATENART_MITTEILUNG);

			var filledAuftrag = elsterTransferService.createBereitstellungsAuftrag(nachricht);

			assertThat(filledAuftrag.getBescheidDatum()).isNull();
		}

		void setup(String datenart) throws IOException {
			when(attachmentService.getCloudFile(anyString())).thenReturn(OzgCloudFileTestFactory.create());
			auftragNeuV3.setDatenart(datenart);
			if (datenart.equals(MukPostfachMessageMapper.DATENART_BESCHEID)) {
				auftragNeuV3.setBescheidDatum(BESCHEID_DATUM);
			}
			when(mapper.toTransferMessage(any())).thenReturn(auftragNeuV3);
			when(restClient.postUpload(any(Path.class))).thenReturn(uploadInfo);
			when(fileMapper.toBereitstellungAnhangWithUploadId(any())).thenReturn(new BereitstellungAnhangWithUploadId());
		}

		@Test
		void shouldNotSetBescheidDatumForKurzmitteilung() {
			auftragNeuV3.setDatenart(MukPostfachMessageMapper.DATENART_KURZMITTEILUNG);
			when(mapper.toTransferMessage(any())).thenReturn(auftragNeuV3);

			var filledAuftrag = elsterTransferService.createBereitstellungsAuftrag(nachricht);

			assertThat(filledAuftrag.getBescheidDatum()).isNull();
		}

		@Test
		void shouldSetAntwortForReplyOptionPossible() {
			PostfachNachricht nachrichtPossible = PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.POSSIBLE)
			  .build();
			auftragNeuV3.setDatenart(MukPostfachMessageMapper.DATENART_KURZMITTEILUNG);
			when(mapper.toTransferMessage(any())).thenReturn(auftragNeuV3);

			var filledAuftrag = elsterTransferService.createBereitstellungsAuftrag(nachrichtPossible);

			assertThat(filledAuftrag.getAntwort()).isNotNull();
		}
	}

	@Nested
	class TestCreatingAntwort {
		@Test
		void shouldCreateMandatoryAntwort() {
			var antwort = elsterTransferService.createAntwort(
			  PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.MANDATORY).build());

			assertThat(antwort).isNotNull();
		}

		@Test
		void shouldCreatePossibleAntwort() {
			var antwort = elsterTransferService.createAntwort(
			  PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.POSSIBLE).build());

			assertThat(antwort).isNotNull();
		}

		@Test
		void shouldNotCreateAntwortWhenForbidden() {
			when(mapper.toTransferMessage(any())).thenReturn(BereitstellungAuftragTestFactory.createMitteilung());

			var auftrag = elsterTransferService.createBereitstellungsAuftrag(
			  PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.FORBIDDEN).build());

			assertThat(auftrag.getAntwort()).isNull();
		}
	}

	@Nested
	class TestSendBescheid {
		@BeforeEach
		void setUp() throws IOException {
			when(attachmentService.getCloudFile(anyString())).thenReturn(OzgCloudFileTestFactory.create());
			when(mapper.toTransferMessage(any())).thenReturn(BereitstellungAuftragTestFactory.createBescheid());
			when(fileMapper.toBereitstellungAnhangWithUploadId(any())).thenReturn(BereitstellungAnhangTestFactory.createWithUploadId());
			when(restClient.postUpload(any(Path.class))).thenReturn(new UploadInfo());
		}

		@Test
		void shouldCallCreateBereitstellungsAuftrag() {
			elsterTransferService.send(PostfachNachrichtTestFactory.create());

			verify(elsterTransferService).createBereitstellungsAuftrag(any(PostfachNachricht.class));
		}

		@Test
		void shouldCallUploadAnhang() {
			elsterTransferService.send(PostfachNachrichtTestFactory.create());

			verify(elsterTransferService).uploadAnhang(anyString(), anyString());
		}

		@Test
		void shouldCallSendBereitstellungsauftrag() {
			elsterTransferService.send(PostfachNachrichtTestFactory.create());

			verify(restClient).postBereitstellungsauftrag(any(BereitstellungAuftragNeuV3.class));
		}

		@Test
		void shouldCallUploadFile() throws IOException {
			elsterTransferService.send(PostfachNachrichtTestFactory.create());

			verify(elsterTransferService).uploadFile(any(OzgCloudFile.class));
		}

		@Test
		void shouldCallAttachmentHandler() {
			elsterTransferService.send(PostfachNachrichtTestFactory.create());

			verify(attachmentService).getFileContent(any(), any());
		}
	}

	@Nested
	class TestSendingMitteilung {
		@BeforeEach
		void setUp() {
			when(mapper.toTransferMessage(any())).thenReturn(BereitstellungAuftragTestFactory.createMitteilung());
		}

		@Test
		void shouldCallSendBereitstellungsauftrag() {
			elsterTransferService.send(PostfachNachrichtTestFactory.createBuilder().attachments(null).build());

			verify(restClient).postBereitstellungsauftrag(any(BereitstellungAuftragNeuV3.class));
		}

		@Test
		void shouldNotCallUploadFile() throws IOException {
			elsterTransferService.send(PostfachNachrichtTestFactory.createBuilder().attachments(null).build());

			verify(elsterTransferService, never()).uploadFile(any(OzgCloudFile.class));
		}
	}

	@Nested
	class TestReadingStatus {
		private static final String ID = "12345678";

		@BeforeEach
		void setUp() {
			when(postfachProperties.getServer()).thenReturn("http:/localhost");
		}

		@Test
		void shouldCallRestTemplate() {
			var url = postfachProperties.getServer() + BEREITSTELLUNGS_ENDPUNKT + "/" + ID;

			elsterTransferService.getStatusOfBereitstellungsauftrag(ID);

			verify(restClient).getForObject(eq(url), any());
		}

		@Test
		void shouldHaveValueNeu() {
			var info = new BereitstellungInfo();
			info.setStatus(BereitstellungAuftragStatus.EINGESTELLT);
			when(restClient.getForObject(anyString(), any())).thenReturn(info);

			var status = elsterTransferService.getStatusOfBereitstellungsauftrag(ID);

			assertThat(status.getStatus()).isEqualTo(BereitstellungAuftragStatus.EINGESTELLT);
		}
	}

	@Nested
	class TestLoadingAntwort {
		byte[] data = "test data".getBytes();

		@Test
		void shouldLoadAntwort() {
			var antwort = elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			assertThat(antwort).isNotNull();
		}

		@Test
		@Disabled("Needs to be reimplemented when the new File API is available")
		void shouldCallDownloadAntwort() {
			setupForAttachments();

			elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			verify(elsterTransferService).downloadAnhang(any());
		}

		@Test
		@Disabled("Needs to be reimplemented when the new File API is available")
		void shouldCallStoreToOzgCloud() throws IOException {
			setupForAttachments();

			elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			verify(elsterTransferService).storeToOzgCloud(any(), any(), any());
		}

		private void setupForAttachments() {
			when(postfachProperties.getServer()).thenReturn("http://localhost");
			when(restClient.getForObject(matches(postfachProperties.getServer() + DOWNLOAD_URI), any())).thenReturn("Test,Data".getBytes());
			when(attachmentService.store(any(), any())).thenReturn(
			  OzgCloudFileId.from(ObjectIdToUUIDConverter.toUUID(new ObjectId().toHexString()).toString()));
		}

		@Test
		@Disabled("Needs to be reimplemented when the new File API is available")
		void shouldHandleAttachmentSaveError() {
			when(attachmentService.store(any(), any())).thenReturn(null);

			DownloadAnhangDetails downloadAnhangDetails = new DownloadAnhangDetails();
			downloadAnhangDetails.setMimeType(OzgCloudFileTestFactory.CONTENT_TYPE);
			downloadAnhangDetails.setDownloadUrl(URI.create("/download"));
			downloadAnhangDetails.setDateiBezeichnung(OzgCloudFileTestFactory.NAME);
			downloadAnhangDetails.setDatei("");

			var res = elsterTransferService.saveAnhang(data, downloadAnhangDetails, new ObjectId(NACHRICHT_ID));

			assertThat(res).isNotPresent();

		}

		@Test
		void shouldHaveMailBody() {
			when(postfachProperties.getBody()).thenReturn(MukPostfachProperties.MAIL_BODY);

			var antwort = elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			assertThat(antwort.getMailBody()).isNotEmpty();
		}

		@Test
		@Disabled("Needs to be reimplemented when the new File API is available")
		void shouldHaveVorgangId() {
			when(restClient.getForObject(matches(postfachProperties.getServer() + DOWNLOAD_URI), any())).thenReturn("Test,Data".getBytes());
			when(attachmentService.store(any(), any())).thenReturn(OzgCloudFileId.from(new ObjectId().toHexString()));

			var antwort = elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			assertThat(antwort.getVorgangId()).isEqualTo(NACHRICHT_ID);
		}

		@Test
		void shouldHaveDirection() {
			var antwort = elsterTransferService.getAntwort(AntwortDetailsTestFactory.create());

			assertThat(antwort.getDirection()).isEqualTo(PostfachNachricht.Direction.IN);
		}
	}

	@Nested
	class TestLoadingAllMessages {
		@BeforeEach
		void setUp() {
			when(postfachProperties.getServer()).thenReturn("http://localhost");
			when(restClient.getForObject(eq(postfachProperties.getServer() + ANTWORTEN_ENDPOINT), any())).thenReturn(
			  AntwortDetailsTestFactory.createAntwortDetailsList());
		}

		@Test
		void shouldCallLoadAntworten() {
			elsterTransferService.getAllMessages();

			verify(elsterTransferService).loadAntworten();
		}

		@Test
		void shouldLoadAntworten() {
			var antwortList = elsterTransferService.getAllMessages();

			assertThat(antwortList).isNotNull();
		}

		@Test
		@Disabled("Saving of Antworten is disabled right now")
		void shouldHavePostfachNachricht() {
			var antwortList = elsterTransferService.getAllMessages().toList();

			assertThat(antwortList).hasSize(1);
			assertThat(antwortList.getFirst()).isInstanceOf(PostfachNachricht.class);
		}
	}

	@Nested
	class TestExceptionHandling {
		byte[] data = "test data".getBytes();

		@Test
		void shouldThrowTechnicalExceptionOnUpload() throws IOException {
			doThrow(IOException.class).when(elsterTransferService).uploadFile(any());
			var id = OzgCloudFileTestFactory.ID.toString();

			assertThatExceptionOfType(TechnicalException.class).isThrownBy(() -> elsterTransferService.uploadAnhang(id, VORGANG_ID.toString()));
		}

		@Test
		void shouldThrowTechnicalExceptionOnStore() throws IOException {
			var downloadAnhangDetails = AntwortDetailsTestFactory.createDownloadAnhangDetails();

			doThrow(IOException.class).when(elsterTransferService).storeToOzgCloud(any(), any(), any());

			assertThatExceptionOfType(TechnicalException.class).isThrownBy(
			  () -> elsterTransferService.saveAnhang(data, downloadAnhangDetails, null)
			);
		}
	}

	@Nested
	class TestCreatingOzgCloudFile {
		DownloadAnhangDetails downloadAnhangDetails = new DownloadAnhangDetails();
		String vorgangId = ObjectId.get().toHexString();

		@BeforeEach
		void setUp() {
			downloadAnhangDetails.setMimeType(OzgCloudFileTestFactory.CONTENT_TYPE);
			downloadAnhangDetails.setDownloadUrl(URI.create("/download"));
			downloadAnhangDetails.setDateiBezeichnung(DATEI_BEZEICHNUNG);
			downloadAnhangDetails.setDatei(OzgCloudFileTestFactory.NAME);
		}

		@Test
		void shouldCreateOzgCloudFile() {
			var file = ElsterTransferRemoteService.createOzgCloudUploadFile(downloadAnhangDetails, vorgangId);

			assertThat(file).isNotNull();
		}

		@Test
		void shouldHaveFileName() {
			var file = ElsterTransferRemoteService.createOzgCloudUploadFile(downloadAnhangDetails, vorgangId);

			assertThat(file.getFileName()).isEqualTo(downloadAnhangDetails.getDatei());
		}

		@Test
		void shouldHaveVorgangId() {
			var file = ElsterTransferRemoteService.createOzgCloudUploadFile(downloadAnhangDetails, vorgangId);

			assertThat(file.getVorgangId()).isEqualTo(vorgangId);
		}

		@Test
		void shouldHaveContentType() {
			var file = ElsterTransferRemoteService.createOzgCloudUploadFile(downloadAnhangDetails, vorgangId);

			assertThat(file.getContentType()).isEqualTo(downloadAnhangDetails.getMimeType());
		}

		@Test
		void shouldHaveFieldName() {
			var file = ElsterTransferRemoteService.createOzgCloudUploadFile(downloadAnhangDetails, vorgangId);

			assertThat(file.getFieldName()).isEqualTo(POSTFACH);
		}

	}

	@Nested
	class TestSaveFile {
		@Test
		void shouldExtractVorgangIdFromDateibezeichnung() throws IOException {
			//when(attachmentService.store(any(), any())).thenReturn(new OzgCloudFileId(UUID.randomUUID().toString()));

			elsterTransferService.saveAnhang("Test,Data".getBytes(), AntwortDetailsTestFactory.createDownloadAnhangDetails(), null);

			verify(elsterTransferService).storeToOzgCloud(any(), eq(VORGANG_OBJECT_ID), any());
		}

		@Test
		void shouldSaveAttachment() {
			//when(attachmentService.store(any(), any())).thenReturn(new OzgCloudFileId(UUID.randomUUID().toString()));

			var fileIdOpt = elsterTransferService.saveAnhang("Test,Data".getBytes(), AntwortDetailsTestFactory.createDownloadAnhangDetails(), null);

			assertThat(fileIdOpt).isNotPresent();
		}
	}
}